/*
 * Copyright (c) 2004, 2005 Christophe Varoqui
 * Copyright (c) 2005 Kiyoshi Ueda, NEC
 * Copyright (c) 2005 Benjamin Marzinski, Redhat
 * Copyright (c) 2005 Edward Goggin, EMC
 */
#include <unistd.h>
#include <sys/stat.h>
#include <libdevmapper.h>
#include <wait.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/mount.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/mount.h>

/*
 * libsysfs
 */
#include <sysfs/libsysfs.h>
#include <sysfs/dlist.h>

/*
 * libcheckers
 */
#include <checkers.h>

/*
 * libmultipath
 */
#include <parser.h>
#include <vector.h>
#include <memory.h>
#include <config.h>
#include <util.h>
#include <hwtable.h>
#include <defaults.h>
#include <structs.h>
#include <callout.h>
#include <blacklist.h>
#include <structs_vec.h>
#include <dmparser.h>
#include <devmapper.h>
#include <dict.h>
#include <discovery.h>
#include <debug.h>
#include <propsel.h>
#include <uevent.h>
#include <switchgroup.h>
#include <print.h>
#include <configure.h>
#include <pgpolicies.h>

#include "main.h"
#include "pidfile.h"
#include "uxlsnr.h"
#include "uxclnt.h"
#include "cli.h"
#include "cli_handlers.h"
#include "lock.h"
#include "waiter.h"
#include "copy.h"

#if !defined(MNT_DETACH)
#define MNT_DETACH 2
#endif

#define FILE_NAME_SIZE 256
#define CMDSIZE 160

#define CALLOUT_DIR "/var/cache/multipathd"

#define LOG_MSG(a,b) \
do { \
	if (strlen(b)) \
		condlog(a, "%s: %s", pp->dev, b); \
} while(0)

pthread_cond_t exit_cond = PTHREAD_COND_INITIALIZER;
pthread_mutex_t exit_mutex = PTHREAD_MUTEX_INITIALIZER;

int exitting = 0;
/*
 * global copy of vecs for use in sig handlers
 */
struct vectors * gvecs;

static int
need_switch_pathgroup (struct multipath * mpp, int refresh)
{
	struct pathgroup * pgp;
	struct path * pp;
	unsigned int i, j;

	if (!mpp || mpp->pgfailback == -FAILBACK_MANUAL)
		return 0;

	/*
	 * Refresh path priority values
	 */
	if (refresh)
		vector_foreach_slot (mpp->pg, pgp, i)
			vector_foreach_slot (pgp->paths, pp, j)
				pathinfo(pp, conf->hwtable, DI_PRIO);

	mpp->bestpg = select_path_group(mpp);

	if (mpp->bestpg != mpp->nextpg)
		return 1;

	return 0;
}

static void
switch_pathgroup (struct multipath * mpp)
{
	mpp->stat_switchgroup++;
	dm_switchgroup(mpp->alias, mpp->bestpg);
	condlog(2, "%s: switch to path group #%i",
		 mpp->alias, mpp->bestpg);
}

static int
coalesce_maps(struct vectors *vecs, vector nmpv)
{
	struct multipath * ompp;
	vector ompv = vecs->mpvec;
	unsigned int i;
	int j;

	vector_foreach_slot (ompv, ompp, i) {
		if (!find_mp_by_wwid(nmpv, ompp->wwid)) {
			/*
			 * remove all current maps not allowed by the
			 * current configuration
			 */
			if (dm_flush_map(ompp->alias, DEFAULT_TARGET)) {
				condlog(0, "%s: unable to flush devmap",
					ompp->alias);
				/*
				 * may be just because the device is open
				 */
				if (!vector_alloc_slot(nmpv))
					return 1;

				vector_set_slot(nmpv, ompp);
				setup_multipath(vecs, ompp);

				if ((j = find_slot(ompv, (void *)ompp)) != -1)
					vector_del_slot(ompv, j);

				continue;
			}
			else {
				dm_lib_release();
				condlog(3, "%s devmap removed", ompp->alias);
			}
		}
	}
	return 0;
}

void
sync_map_state(struct multipath *mpp)
{
	struct pathgroup *pgp;
        struct path *pp;
	unsigned int i, j;

	vector_foreach_slot (mpp->pg, pgp, i){
		vector_foreach_slot (pgp->paths, pp, j){
			if (pp->state <= PATH_UNCHECKED)
				continue;
			if ((pp->dmstate == PSTATE_FAILED ||
			     pp->dmstate == PSTATE_UNDEF) &&
			    (pp->state == PATH_UP || pp->state == PATH_GHOST))
				dm_reinstate_path(mpp->alias, pp->dev_t);
			else if ((pp->dmstate == PSTATE_ACTIVE ||
				  pp->dmstate == PSTATE_UNDEF) &&
				 (pp->state == PATH_DOWN ||
				  pp->state == PATH_SHAKY))
				dm_fail_path(mpp->alias, pp->dev_t);
		}
	}
}

static void
sync_maps_state(vector mpvec)
{
	unsigned int i;
	struct multipath *mpp;

	vector_foreach_slot (mpvec, mpp, i) 
		sync_map_state(mpp);
}

static int
flush_map(struct multipath * mpp, struct vectors * vecs)
{
	/*
	 * clear references to this map before flushing so we can ignore
	 * the spurious uevent we may generate with the dm_flush_map call below
	 */
	if (dm_flush_map(mpp->alias, DEFAULT_TARGET)) {
		/*
		 * May not really be an error -- if the map was already flushed
		 * from the device mapper by dmsetup(8) for instance.
		 */
		condlog(0, "%s: can't flush", mpp->alias);
		return 1;
	}
	else {
		dm_lib_release();
		condlog(3, "%s: devmap removed", mpp->alias);
	}

	orphan_paths(vecs->pathvec, mpp);
	remove_map(mpp, vecs, stop_waiter_thread, 1);

	return 0;
}

static int
uev_add_map (char * devname, struct vectors * vecs)
{
	condlog(2, "%s: add map (uevent)", devname);
	ev_add_map(devname, vecs);
	return 0;
}

int
ev_add_map (char * devname, struct vectors * vecs)
{
	int major, minor;
	char dev_t[BLK_DEV_SIZE];
	char * alias;
	char * refwwid;
	struct multipath * mpp;
	int map_present;
	int r = 1;

	/* libsysfs seems to forget to terminate the string... */
	memset(dev_t, 0, BLK_DEV_SIZE);
	if (sscanf(devname, "dm-%d", &minor) == 1 &&
	    !sysfs_get_dev(sysfs_path, devname, dev_t, BLK_DEV_SIZE) &&
	    sscanf(dev_t, "%d:%d", &major, &minor) == 2)
		alias = dm_mapname(major, minor);
	else
		alias = STRDUP(devname);
		
	if (!alias)
		return 1;
	
	map_present = dm_map_present(alias);

	if (map_present && dm_type(alias, DEFAULT_TARGET) <= 0) {
		condlog(4, "%s: not a multipath map", alias);
		FREE(alias);
		return 0;
	}

	mpp = find_mp_by_alias(vecs->mpvec, alias);

	if (mpp) {
                /*
		 * Not really an error -- we generate our own uevent
		 * if we create a multipath mapped device as a result
		 * of uev_add_path
		 */
		condlog(0, "%s: devmap already registered",
			devname);
		FREE(alias);
		return 0;
	}

	/*
	 * now we can register the map
	 */
	if (map_present && (mpp = add_map_without_path(vecs, minor, alias,
					start_waiter_thread))) {
		sync_map_state(mpp);
		condlog(3, "%s: devmap %s added", alias, devname);
		return 0;
	}
	refwwid = get_refwwid(devname, DEV_DEVMAP, vecs->pathvec);

	if (refwwid) {
		r = coalesce_paths(vecs, NULL, refwwid);
		dm_lib_release();
	}
	
	if (!r)
		condlog(3, "%s: devmap %s added", alias, devname);
	else
		condlog(0, "%s: uev_add_map %s failed", alias, devname);

	FREE(refwwid);
	FREE(alias);
	return r;
}

static int
uev_remove_map (char * devname, struct vectors * vecs)
{
	struct multipath * mpp;

	condlog(2, "%s: remove map (uevent)", devname);

	mpp = find_mp_by_str(vecs->mpvec, devname);
	if (!mpp) {
		condlog(2, "%s: devmap not registered, can't remove", devname);
		return 0;
	}
	orphan_paths(vecs->pathvec, mpp);
	remove_map(mpp, vecs, stop_waiter_thread, 1);
	return 0;
}

int
ev_remove_map (char * devname, struct vectors * vecs)
{
	struct multipath * mpp;

	mpp = find_mp_by_str(vecs->mpvec, devname);

	if (!mpp) {
		condlog(3, "%s: devmap not registered, can't remove",
			devname);
		return 0;
	}
	flush_map(mpp, vecs);

	return 0;
}

static int
uev_umount_map (char * devname, struct vectors * vecs)
{
	struct multipath * mpp;

	condlog(2, "%s: umount map (uevent)", devname);

	mpp = find_mp_by_str(vecs->mpvec, devname);

	if (!mpp)
		return 0;

	update_mpp_paths(mpp, vecs->pathvec);
	verify_paths(mpp, vecs, NULL);

	if (!VECTOR_SIZE(mpp->paths))
		flush_map(mpp, vecs);

	return 0;
}
	
static int
uev_add_path (char * devname, struct vectors * vecs)
{
	condlog(2, "%s: add path (uevent)", devname);
	return (ev_add_path(devname, vecs) != 1)? 0 : 1;
}


/*
 * returns:
 * 0: added
 * 1: error
 * 2: blacklisted
 */
int
ev_add_path (char * devname, struct vectors * vecs)
{
	struct multipath * mpp;
	struct path * pp;
	char empty_buff[WWID_SIZE] = {0};
	int ret;
	int retrys = 10;

	ret = device_ok_to_add(devname);
	if (ret < 0)
		return 1;
	if (ret == 0)
		/* Is 2 the best return code? */
		return 2;

	pp = find_path_by_dev(vecs->pathvec, devname);

	if (pp) {
		condlog(0, "%s: spurious uevent, path already in pathvec",
			devname);
		if (pp->mpp)
			return 0;
	}
	else {
		/*
		 * get path vital state
		 */
		if (!(pp = store_pathinfo(vecs->pathvec, conf->hwtable,
		      devname, DI_ALL))) {
			condlog(0, "%s: failed to store path info", devname);
			return 1;
		}
		pp->checkint = conf->checkint;
	}

	/*
	 * need path UID to go any further
	 */
	if (memcmp(empty_buff, pp->wwid, WWID_SIZE) == 0) {
		condlog(0, "%s: failed to get path uid", devname);
		return 1; /* leave path added to pathvec */
	}
	if (blacklist_path(conf, pp)){
		int i = find_slot(vecs->pathvec, (void *)pp);
		if (i != -1)
			vector_del_slot(vecs->pathvec, i);
		free_path(pp);
		return 2;
	}	
	mpp = pp->mpp = find_mp_by_wwid(vecs->mpvec, pp->wwid);
rescan:
	if (mpp) {
		if (adopt_paths(vecs->pathvec, mpp))
			return 1; /* leave path added to pathvec */

		mpp->hwe = pp->hwe;
		verify_paths(mpp, vecs, NULL);
		mpp->flush_on_last_del = FLUSH_UNDEF;
		mpp->action = ACT_RELOAD;
	}
	else {
		if ((mpp = add_map_with_path(vecs, pp, 1)))
			mpp->action = ACT_CREATE;
		else
			return 1; /* leave path added to pathvec */
	}

	/*
	 * push the map to the device-mapper
	 */
	if (setup_map(mpp)) {
		condlog(0, "%s: failed to setup map for addition of new "
			"path %s", mpp->alias, devname);
		goto out;
	}
	/*
	 * reload the map for the multipath mapped device
	 */
	if (domap(mpp) <= 0) {
		condlog(0, "%s: failed in domap for addition of new "
			"path %s", mpp->alias, devname);
		/*
 		 * deal with asynchronous uevents :((
 		 */
		if (mpp->action == ACT_RELOAD && retrys > 0) {
			condlog(0, "%s: uev_add_path sleep", mpp->alias);
			sleep(1);
			retrys--;
			update_mpp_paths(mpp, vecs->pathvec);
			goto rescan;
		}
		else
			goto out;
	}
	dm_lib_release();

	/*
	 * update our state from kernel regardless of create or reload
	 */
	if (setup_multipath(vecs, mpp))
		goto out;

	sync_map_state(mpp);

	if (mpp->action == ACT_CREATE &&
	    start_waiter_thread(mpp, vecs))
			goto out;

	condlog(3, "%s path added to devmap %s", devname, mpp->alias);
	return 0;

out:
	remove_map(mpp, vecs, NULL, 1);
	return 1;
}

static int
uev_remove_path (char * devname, struct vectors * vecs)
{
	condlog(2, "%s: remove path (uevent)", devname);
	return ev_remove_path(devname, vecs);
}

int
ev_remove_path (char * devname, struct vectors * vecs)
{
	struct multipath * mpp;
	struct path * pp;
	int i;
	int rm_path = 1;

	pp = find_path_by_dev(vecs->pathvec, devname);

	if (!pp) {
		condlog(0, "%s: spurious uevent, path not in pathvec", devname);
		return 1;
	}

	/*
	 * avoid referring to the map of an orphanned path
	 */
	if ((mpp = pp->mpp)) {

		/*
		 * remove the map IFF removing the last path
		 */
		if (pathcount(mpp, PATH_WILD) > 1) {
			vector rpvec = vector_alloc();

			/*
	 	 	 * transform the mp->pg vector of vectors of paths
	 	 	 * into a mp->params string to feed the device-mapper
	 	 	 */
			update_mpp_paths(mpp, vecs->pathvec);
			if ((i = find_slot(mpp->paths, (void *)pp)) != -1)
				vector_del_slot(mpp->paths, i);

			if (VECTOR_SIZE(mpp->paths) == 0) {
				char alias[WWID_SIZE];

				/*
				 * flush_map will fail if the device is open
				 */
				strncpy(alias, mpp->alias, WWID_SIZE);
				if (mpp->flush_on_last_del == FLUSH_ENABLED) {
					condlog(2, "%s Last path deleted, disabling queueing", mpp->alias);
					mpp->retry_tick = 0;
					mpp->no_path_retry = NO_PATH_RETRY_FAIL;
					mpp->flush_on_last_del = FLUSH_IN_PROGRESS;
					dm_queue_if_no_path(mpp->alias, 0);
				}
				if (flush_map(mpp, vecs))
					rm_path = 0;
				else
					condlog(3, "%s: removed map after removing"
						" multiple paths", alias);
			}
			else {
				if (setup_map(mpp)) {
					condlog(0, "%s: failed to setup map for"
						" removal of path %s", mpp->alias, devname);
					free_pathvec(rpvec, KEEP_PATHS);
					goto out;
				}
				/*
	 	 	 	 * reload the map
	 	 	 	 */
				mpp->action = ACT_RELOAD;
				if (domap(mpp) <= 0) {
					condlog(0, "%s: failed in domap for "
						"removal of path %s",
						mpp->alias, devname);
					/*
					 * Delete path from pathvec so that
					 * update_mpp_paths wont find it later
					 * when/if another path is removed.
					 */
					if ((i = find_slot(vecs->pathvec, (void *)pp)) != -1)
						vector_del_slot(vecs->pathvec, i);
					free_path(pp);
					return 1;
				}
				/*
				 * update our state from kernel
				 */
				if (setup_multipath(vecs, mpp)) {
					free_pathvec(rpvec, KEEP_PATHS);
					goto out;
				}
				sync_map_state(mpp);

				condlog(3, "%s: path removed from map %s",
					devname, mpp->alias);
			}
			free_pathvec(rpvec, KEEP_PATHS);
		}
		else {
			char alias[WWID_SIZE];

			/*
			 * flush_map will fail if the device is open
			 */
			strncpy(alias, mpp->alias, WWID_SIZE);
			if (mpp->flush_on_last_del == FLUSH_ENABLED) {
				condlog(2, "%s Last path deleted, disabling queueing", mpp->alias);
				mpp->retry_tick = 0;
				mpp->no_path_retry = NO_PATH_RETRY_FAIL;
				mpp->flush_on_last_del = FLUSH_IN_PROGRESS;
				dm_queue_if_no_path(mpp->alias, 0);
			}
			if (flush_map(mpp, vecs))
				rm_path = 0;
			else
				condlog(3, "%s: removed map", alias);
		}
	}

	if (rm_path) {
		if ((i = find_slot(vecs->pathvec, (void *)pp)) != -1)
			vector_del_slot(vecs->pathvec, i);
		free_path(pp);
	}

	return 0;

out:
	remove_map(mpp, vecs, stop_waiter_thread, 1);
	return 1;
}

static int
map_discovery (struct vectors * vecs)
{
	struct multipath * mpp;
	unsigned int i;

	if (dm_get_maps(vecs->mpvec, "multipath"))
		return 1;

	vector_foreach_slot (vecs->mpvec, mpp, i)
		if (setup_multipath(vecs, mpp))
			return 1;

	return 0;
}

int
uxsock_trigger (char * str, char ** reply, int * len, void * trigger_data)
{
	struct vectors * vecs;
	int r;
	
	*reply = NULL;
	*len = 0;
	vecs = (struct vectors *)trigger_data;

	pthread_cleanup_push(cleanup_lock, vecs->lock);
	lock(vecs->lock);
	pthread_testcancel();

	r = parse_cmd(str, reply, len, vecs);

	if (r > 0) {
		*reply = STRDUP("fail\n");
		*len = strlen(*reply) + 1;
		r = 1;
	}
	else if (!r && *len == 0) {
		*reply = STRDUP("ok\n");
		*len = strlen(*reply) + 1;
		r = 0;
	}
	/* else if (r < 0) leave *reply alone */

	lock_cleanup_pop(vecs->lock);
	return r;
}

static int
uev_discard(char * devpath)
{
	char a[11], b[11];

	/*
	 * keep only block devices, discard partitions
	 */
	if (sscanf(devpath, "/block/%10s", a) != 1 ||
	    sscanf(devpath, "/block/%10[^/]/%10s", a, b) == 2) {
		condlog(4, "discard event on %s", devpath);
		return 1;
	}
	return 0;
}

int 
uev_trigger (struct uevent * uev, void * trigger_data)
{
	int r = 0;
	char devname[32];
	struct vectors * vecs;

	vecs = (struct vectors *)trigger_data;

	if (uev_discard(uev->devpath))
		return 0;

	basenamecpy(uev->devpath, devname);
	pthread_cleanup_push(cleanup_lock, vecs->lock);
	lock(vecs->lock);
	pthread_testcancel();

	/*
	 * device map event
	 * Add events are ignored here as the tables
	 * are not fully initialised then.
	 */
	if (!strncmp(devname, "dm-", 3)) {
		if (!strncmp(uev->action, "change", 6)) {
			r = uev_add_map(devname, vecs);
			goto out;
		}
		if (!strncmp(uev->action, "remove", 6)) {
			r = uev_remove_map(devname, vecs);
			goto out;
		}
		if (!strncmp(uev->action, "umount", 6)) {
			r = uev_umount_map(devname, vecs);
			goto out;
		}
		goto out;
	}
	
	/*
	 * path add/remove event
	 */
	if (blacklist(conf->blist_devnode, conf->elist_devnode, devname))
		goto out;

	if (!strncmp(uev->action, "add", 3)) {
		r = uev_add_path(devname, vecs);
		goto out;
	}
	if (!strncmp(uev->action, "remove", 6)) {
		r = uev_remove_path(devname, vecs);
		goto out;
	}

out:;
	lock_cleanup_pop(vecs->lock);
	return r;
}

static void *
ueventloop (void * ap)
{
	block_signal(SIGUSR1, NULL);
	block_signal(SIGHUP, NULL);

	if (uevent_listen(&uev_trigger, ap))
		fprintf(stderr, "error starting uevent listener");
		
	return NULL;
}

static void *
uxlsnrloop (void * ap)
{
	if (load_keys())
		return NULL;
	
	if (alloc_handlers())
		return NULL;

	block_signal(SIGUSR1, NULL);
	block_signal(SIGHUP, NULL);

	add_handler(LIST+PATHS, cli_list_paths);
	add_handler(LIST+MAPS, cli_list_maps);
	add_handler(LIST+MAPS+STATUS, cli_list_maps_status);
	add_handler(LIST+MAPS+STATS, cli_list_maps_stats);
	add_handler(LIST+MAPS+TOPOLOGY, cli_list_maps_topology);
	add_handler(LIST+TOPOLOGY, cli_list_maps_topology);
	add_handler(LIST+MAP+TOPOLOGY, cli_list_map_topology);
	add_handler(LIST+CONFIG, cli_list_config);
	add_handler(LIST+BLACKLIST, cli_list_blacklist);
	add_handler(LIST+DEVICES, cli_list_devices);
	add_handler(ADD+PATH, cli_add_path);
	add_handler(DEL+PATH, cli_del_path);
	add_handler(ADD+MAP, cli_add_map);
	add_handler(DEL+MAP, cli_del_map);
	add_handler(SWITCH+MAP+GROUP, cli_switch_group);
	add_handler(RECONFIGURE, cli_reconfigure);
	add_handler(SUSPEND+MAP, cli_suspend);
	add_handler(RESUME+MAP, cli_resume);
	add_handler(REINSTATE+PATH, cli_reinstate);
	add_handler(FAIL+PATH, cli_fail);
	add_handler(FORCEQ+MPATH_DAEMON, cli_force_no_daemon_q);
	add_handler(RESTOREQ+MPATH_DAEMON, cli_restore_no_daemon_q);
	add_handler(DISABLEQ+MAP, cli_disable_queueing);
	add_handler(RESTOREQ+MAP, cli_restore_queueing);
	add_handler(DISABLEQ+MAPS, cli_disable_all_queueing);
	add_handler(RESTOREQ+MAPS, cli_restore_all_queueing);
	add_handler(RESIZE+MAP, cli_resize);

	umask(077);
	uxsock_listen(&uxsock_trigger, ap);

	return NULL;
}

static int
exit_daemon (int status)
{
	if (status != 0)
		fprintf(stderr, "bad exit status. see daemon.log\n");

	condlog(3, "umount ramfs");
	umount(CALLOUT_DIR);

	condlog(3, "unlink pidfile");
	unlink(DEFAULT_PIDFILE);

	lock(&exit_mutex);
	pthread_cond_signal(&exit_cond);
	unlock(&exit_mutex);

	return status;
}

static void
fail_path (struct path * pp, int del_active)
{
	if (!pp->mpp)
		return;

	condlog(2, "checker failed path %s in map %s",
		 pp->dev_t, pp->mpp->alias);

	dm_fail_path(pp->mpp->alias, pp->dev_t);
	if (del_active)
		update_queue_mode_del_path(pp->mpp);
}

/*
 * caller must have locked the path list before calling that function
 */
static void
reinstate_path (struct path * pp, int add_active)
{
	if (!pp->mpp)
		return;

	if (dm_reinstate_path(pp->mpp->alias, pp->dev_t))
		condlog(0, "%s: reinstate failed", pp->dev_t);
	else {
		condlog(2, "%s: reinstated", pp->dev_t);
		if (add_active)
			update_queue_mode_add_path(pp->mpp);
	}
}

static void
enable_group(struct path * pp)
{
	struct pathgroup * pgp;

	/*
	 * if path is added through uev_add_path, pgindex can be unset.
	 * next update_strings() will set it, upon map reload event.
	 *
	 * we can safely return here, because upon map reload, all
	 * PG will be enabled.
 	 */
	if (!pp->mpp->pg || !pp->pgindex)
		return;

	pgp = VECTOR_SLOT(pp->mpp->pg, pp->pgindex - 1);
	
	if (pgp->status == PGSTATE_DISABLED) {
		condlog(2, "%s: enable group #%i", pp->mpp->alias, pp->pgindex);
		dm_enablegroup(pp->mpp->alias, pp->pgindex);
	}
}

static void
mpvec_garbage_collector (struct vectors * vecs)
{
	struct multipath * mpp;
	unsigned int i;

	vector_foreach_slot (vecs->mpvec, mpp, i) {
		if (mpp && mpp->alias && !dm_map_present(mpp->alias)) {
			condlog(2, "%s: remove dead map", mpp->alias);
			remove_map(mpp, vecs, stop_waiter_thread, 1);
			i--;
		}
	}
}

static void
defered_failback_tick (vector mpvec)
{
	struct multipath * mpp;
	unsigned int i;

	vector_foreach_slot (mpvec, mpp, i) {
		/*
		 * defered failback getting sooner
		 */
		if (mpp->pgfailback > 0 && mpp->failback_tick > 0) {
			mpp->failback_tick--;

			if (!mpp->failback_tick && need_switch_pathgroup(mpp, 1))
				switch_pathgroup(mpp);
		}
	}
}

static void
retry_count_tick(vector mpvec)
{
	struct multipath *mpp;
	unsigned int i;

	vector_foreach_slot (mpvec, mpp, i) {
		if (mpp->retry_tick) {
			mpp->stat_total_queueing_time++;
			condlog(4, "%s: Retrying.. No active path", mpp->alias);
			if(--mpp->retry_tick == 0) {
				dm_queue_if_no_path(mpp->alias, 0);
				condlog(2, "%s: Disable queueing", mpp->alias);
			}
		}
	}
}

int
check_sysfs_state (struct path *pp, int *newstate)
{
	char state[32];
	if (pp->bus != SYSFS_BUS_SCSI)
		return 0;
	if (sysfs_get_state(sysfs_path, pp->dev, state, 32) != 0)
                return 0;
	condlog(3, "%s: state = %s", pp->dev, state);
	if (strncmp(state, "blocked", 7) == 0){
		*newstate = PATH_PENDING;
		return 1;
	}
	return 0;
}

int update_prio(struct path *pp, int refresh_all)
{
	int oldpriority;
	struct pathgroup * pgp;
	struct path *pp1;
	int i, j, changed = 0;

	if (refresh_all) {
		vector_foreach_slot (pp->mpp->pg, pgp, i) {
			vector_foreach_slot (pgp->paths, pp1, j) {
				oldpriority = pp1->priority;
				pathinfo(pp1, conf->hwtable, DI_PRIO);
				if (pp1->priority != oldpriority)
					changed = 1;
			}
		}
		return changed;
	}
	oldpriority = pp->priority;
	pathinfo(pp, conf->hwtable, DI_PRIO);

	if (pp->priority == oldpriority)
		return 0;
	return 1;
}

int update_path_groups(struct multipath *mpp, struct vectors *vecs, int refresh)
{
	int i;
	struct path * pp;

	update_mpp_paths(mpp, vecs->pathvec);
	if (refresh) {
		vector_foreach_slot (mpp->paths, pp, i)
			pathinfo(pp, conf->hwtable, DI_PRIO);
	}
	setup_map(mpp);
	mpp->action = ACT_RELOAD;
	if (domap(mpp) <= 0) {
		condlog(0, "%s: failed to update map : %s", mpp->alias,
				strerror(errno));
		return 1;
	}
	dm_lib_release();
	if (setup_multipath(vecs, mpp) != 0)
		return 1;
	sync_map_state(mpp);

	return 0;
}


static void *
checkerloop (void *ap)
{
	struct vectors *vecs;
	struct path *pp;
	int count = 0;
	int newstate;
	unsigned int i;
	sigset_t old;

	mlockall(MCL_CURRENT | MCL_FUTURE);
	vecs = (struct vectors *)ap;
	condlog(2, "path checkers start up");

	/*
	 * init the path check interval
	 */
	vector_foreach_slot (vecs->pathvec, pp, i) {
		pp->checkint = conf->checkint;
	}

	while (1) {
		block_signal(SIGHUP, &old);
		pthread_cleanup_push(cleanup_lock, vecs->lock);
		lock(vecs->lock);
		pthread_testcancel();
		condlog(4, "tick");

		vector_foreach_slot (vecs->pathvec, pp, i) {
			int new_path_up = 0;
			if (!pp->mpp)
				continue;

			if (pp->tick && --pp->tick)
				continue; /* don't check this path yet */

			/*
			 * provision a next check soonest,
			 * in case we exit abnormaly from here
			 */
			pp->tick = conf->checkint;
			
			if (!checker_selected(&pp->checker)) {
				pathinfo(pp, conf->hwtable, DI_SYSFS);
				select_checker(pp);
			}

			if (!checker_selected(&pp->checker)) {
				condlog(0, "%s: checker is not set", pp->dev);
				continue;
			}

			/*
			 * Set checker in async mode.
			 * Honored only by checker implementing the said mode.
			 */
			checker_set_async(&pp->checker);

			if (check_sysfs_state(pp, &newstate) == 0) {
				if (!conf->checker_timeout)
					sysfs_get_timeout(sysfs_path, pp->dev,
							&(pp->checker.timeout));
				newstate = checker_check(&pp->checker);
			}
			
			if (newstate < 0) {
				condlog(2, "%s: unusable path", pp->dev);
				pathinfo(pp, conf->hwtable, 0);
				continue;
			}

			/*
			 * Async IO in flight. Keep the previous path state
			 * and reschedule as soon as possible.
			 */
			if (newstate == PATH_PENDING) {
				condlog(4, "%s: pending path", pp->dev);
				pp->tick = 1;
				continue;
			}

			if (newstate != pp->state) {
				int oldstate = pp->state;
				pp->state = newstate;
				LOG_MSG(1, checker_message(&pp->checker));

				/*
				 * upon state change, reset the checkint
				 * to the shortest delay
				 */
				pp->checkint = conf->checkint;

				if (newstate == PATH_DOWN ||
				    newstate == PATH_SHAKY ||
				    update_multipath_strings(pp->mpp,
					    		     vecs->pathvec)) {
					/*
					 * proactively fail path in the DM
					 */
					if (oldstate == PATH_UP ||
					    oldstate == PATH_GHOST)
						fail_path(pp, 1);
					else
						fail_path(pp, 0);

					/*
					 * cancel scheduled failback
					 */
					pp->mpp->failback_tick = 0;

					pp->mpp->stat_path_failures++;
					continue;
				}

				/*
				 * reinstate this path
				 */
				if (oldstate != PATH_UP &&
				    oldstate != PATH_GHOST)
					reinstate_path(pp, 1);
				else
					reinstate_path(pp, 0);
				new_path_up = 1;

				/*
				 * if at least one path is up in a group, and
				 * the group is disabled, re-enable it
				 */
				if (newstate == PATH_UP)
					enable_group(pp);
			}
			else if (newstate == PATH_UP || newstate == PATH_GHOST) {
				LOG_MSG(4, checker_message(&pp->checker));
				/*
				 * double the next check delay.
				 * max at conf->max_checkint
				 */
				if (pp->checkint < (conf->max_checkint / 2))
					pp->checkint = 2 * pp->checkint;
				else
					pp->checkint = conf->max_checkint;

				pp->tick = pp->checkint;
				condlog(4, "%s: delay next check %is",
						pp->dev_t, pp->tick);
			}
			else if (newstate == PATH_DOWN) {
				if (conf->log_checker_err == LOG_CHKR_ERR_ONCE)
					LOG_MSG(3, checker_message(&pp->checker));
				else
					LOG_MSG(2, checker_message(&pp->checker));
			}

			pp->state = newstate;

			/*
			 * path prio refreshing
			 */
			condlog(4, "path prio refresh");
			if (update_prio(pp, new_path_up) && pp->mpp->pgpolicyfn == (pgpolicyfn *)group_by_prio && (pp->mpp->pgfailback == -FAILBACK_IMMEDIATE || pp->mpp->pgfailback > 0))
				update_path_groups(pp->mpp, vecs, !new_path_up);
			else if (need_switch_pathgroup(pp->mpp, 0)) {
				if (pp->mpp->pgfailback > 0 &&
				    (new_path_up ||
				     pp->mpp->failback_tick <= 0))
					pp->mpp->failback_tick =
						pp->mpp->pgfailback + 1;
				else if (pp->mpp->pgfailback ==
						-FAILBACK_IMMEDIATE)
					switch_pathgroup(pp->mpp);
			}
		}
		defered_failback_tick(vecs->mpvec);
		retry_count_tick(vecs->mpvec);

		if (count)
			count--;
		else {
			condlog(4, "map garbage collection");
			mpvec_garbage_collector(vecs);
			count = MAPGCINT;
		}
		
		lock_cleanup_pop(vecs->lock);
		pthread_sigmask(SIG_SETMASK, &old, NULL);
		sleep(1);
	}
	return NULL;
}

static void
orphan_unused_paths (struct vectors *vecs)
{
	int i;
	struct path *pp;

	vector_foreach_slot(vecs->pathvec, pp, i) {
		if (pp->mpp == NULL)
			orphan_path(pp);
	}
}

int
configure (struct vectors * vecs, int start_waiters)
{
	struct multipath * mpp;
	struct path * pp;
	vector mpvec;
	int i;

	if (!vecs->pathvec && !(vecs->pathvec = vector_alloc()))
		return 1;
	
	if (!vecs->mpvec && !(vecs->mpvec = vector_alloc()))
		return 1;
	
	if (!(mpvec = vector_alloc()))
		return 1;

	/*
	 * probe for current path (from sysfs) and map (from dm) sets
	 */
	path_discovery(vecs->pathvec, conf, DI_ALL);

	vector_foreach_slot (vecs->pathvec, pp, i){
		if (blacklist_path(conf, pp)){
			vector_del_slot(vecs->pathvec, i);
			free_path(pp);
			i--;
		}	
		else
			pp->checkint = conf->checkint;
	}
	if (map_discovery(vecs))
		return 1;

	/*
	 * create new set of maps & push changed ones into dm
	 */
	if (coalesce_paths(vecs, mpvec, NULL))
		return 1;

	/*
	 * may need to remove some maps which are no longer relevant
	 * e.g., due to blacklist changes in conf file
	 */
	if (coalesce_maps(vecs, mpvec))
		return 1;

	dm_lib_release();

	sync_maps_state(mpvec);

	if (conf->verbosity > 2)
		vector_foreach_slot(mpvec, mpp, i)
			print_map(mpp);

	/*
	 * purge dm of old maps
	 */
	remove_maps(vecs, NULL);

	/*
	 * save new set of maps formed by considering current path state
	 */
	vector_free(vecs->mpvec);
	vecs->mpvec = mpvec;

	orphan_unused_paths(vecs);
	/*
	 * start dm event waiter threads for these new maps
	 */
	vector_foreach_slot(vecs->mpvec, mpp, i) {
		if (setup_multipath(vecs, mpp))
			return 1;
		if (start_waiters)
			if (start_waiter_thread(mpp, vecs))
				return 1;
	}
	return 0;
}

int
reconfigure (struct vectors * vecs)
{
	struct config * old = conf;

	/*
	 * free old map and path vectors ... they use old conf state
	 */
	if (VECTOR_SIZE(vecs->mpvec))
		remove_maps(vecs, stop_waiter_thread);

	if (VECTOR_SIZE(vecs->pathvec))
		free_pathvec(vecs->pathvec, FREE_PATHS);

	vecs->pathvec = NULL;
	conf = NULL;

	if (load_config(DEFAULT_CONFIGFILE))
		return 1;

	conf->verbosity = old->verbosity;

	if (!conf->checkint) {
		conf->checkint = DEFAULT_CHECKINT;
		conf->max_checkint = MAX_CHECKINT(conf->checkint);
	}
	configure(vecs, 1);
	free_config(old);
	return 0;
}

static struct vectors *
init_vecs (void)
{
	struct vectors * vecs;

	vecs = (struct vectors *)MALLOC(sizeof(struct vectors));

	if (!vecs)
		return NULL;

	vecs->lock = 
		(pthread_mutex_t *)MALLOC(sizeof(pthread_mutex_t));

	if (!vecs->lock)
		goto out;

	pthread_mutex_init(vecs->lock, NULL);

	return vecs;

out:
	FREE(vecs);
	condlog(0, "failed to init paths");
	return NULL;
}

static int
find_keep_dirs(void)
{
	int i, j, r;
	char link_target[PATH_MAX];
	char *dir;
	char *needed_dirs[15] = { "/sbin", "/bin", "/tmp", "/usr", "/usr/lib",
				  "/etc", "/proc", "/sys", "/var", "/var/lib",
				  "/var/run", "/lib", "/lib64", "/usr/lib64",
				  "/ram" };
	for (i = 0; i < 15; i++) {
		r = readlink(needed_dirs[i], link_target, PATH_MAX - 1);
		if (r <= 0)
			continue;
		link_target[r] = '\0';
		vector_foreach_slot(conf->keep_dirs, dir, j)
			if (strcmp(dir, link_target) == 0)
				goto next;
		dir = strdup(link_target);
		if (!dir) {
			condlog(0, "can't allocate symlink space for %s",
				needed_dirs[i]);
			return -1;
		}
		if (!vector_alloc_slot(conf->keep_dirs)) {
			condlog(0, "can't store symlink for %s",
				needed_dirs[i]);
			free(dir);
			return -1;
		}
		vector_set_slot(conf->keep_dirs, dir);
next:
		;
	}
	return 0;
}

static int
unmount_extra_devs(void)
{
	char buf[LINE_MAX];
	FILE *file;
	vector extra_devs;
	char *slot;
	int s;

	file = fopen("/proc/mounts", "r");
	if (!file) {
		condlog(0, "couldn't open /proc/mounts : %s", strerror(errno));
		return -1;
	}

	extra_devs = vector_alloc();
	if (!extra_devs) {
		condlog(0, "couldn't allocate vector for extra devs\n");
		return -1;
	}

	find_keep_dirs();
	while (fgets(buf, LINE_MAX, file)) {
		int i;
		char *keep;
		char *end, *extra, *mnt = strchr(buf, ' ');
		if (!mnt)
			continue;
		mnt++;
		end = strchr(mnt, ' ');
		if (!end) {
			condlog(2, "can't process /proc/mounts line (%s). skipping", buf);
			continue;
		}
		*end = '\0';
		if (conf->keep_dirs) {
			vector_foreach_slot(conf->keep_dirs, keep, i) {
				int mnt_len = strlen(mnt);
				int keep_len = strlen(keep);
				if (strncmp(mnt, keep, mnt_len) == 0 &&
				    (mnt_len == keep_len ||
				     (mnt_len < keep_len &&
				      keep[mnt_len] == '/')))
					goto next;
			}
		}
		if (strcmp(mnt, "/") == 0 || strcmp(mnt, "/sbin") == 0 ||
		    strcmp(mnt, "/bin") == 0 || strcmp(mnt, "/tmp") == 0 ||
		    strcmp(mnt, "/usr") == 0 || strncmp(mnt, "/usr/lib", 8) == 0 ||
		    strcmp(mnt, "/etc") == 0 || strncmp(mnt, "/proc", 5) == 0 ||
		    strcmp(mnt, "/dev") == 0 || strncmp(mnt, "/sys", 4) == 0 ||
		    strcmp(mnt, "/var") == 0 || strcmp(mnt, "/var/lib") == 0 ||
		    strcmp(mnt, "/var/lib/multipath") == 0 ||
		    strncmp(mnt, "/dev/", 5) == 0 || strcmp(mnt, CALLOUT_DIR) == 0 ||
		    strncmp(mnt, "/var/run", 8) == 0 || strncmp(mnt, "/lib", 4) == 0 ||
		    strncmp(mnt, "/lib64", 6) == 0 || strncmp(mnt, "/usr/lib64", 10) == 0 ||
		    strncmp(mnt, "/ram", 4) == 0 || strcmp(mnt, "/var/cache") == 0)
			continue;
		if (!vector_alloc_slot(extra_devs)) {
			condlog(2, "couldn't allocate vector slot for extra dev (%s). "
				"skipping,", mnt);
			continue;
 		}
		extra = strdup(mnt);
		vector_set_slot(extra_devs, extra);
next:
		;
	}
	fclose(file);

	vector_foreach_slot(extra_devs, slot, s) {
	        if (umount2(slot, MNT_DETACH) < 0 && errno != ENOENT) {
                        condlog(0, "failed to umount '%s' (%s). skipping",
				slot, strerror(errno));
                }
		free(slot);
	}
	vector_free(extra_devs);
	return 0;
}	
	
	


#ifdef CLONE_NEWNS
static int
prepare_namespace(void)
{
	mode_t mode = S_IRWXU;
	struct stat * buf;
	char ramfs_args[64];
	int i;
	int fd;
	char * bin;
	size_t size = 10;
	struct stat statbuf;

	if (!conf->binvec)
		return 0;
	buf = (struct stat *)MALLOC(sizeof(struct stat));

	/*
	 * create a temp mount point for ramfs
	 */
	if (stat(CALLOUT_DIR, buf) < 0) {
		if (mkdir(CALLOUT_DIR, mode) < 0) {
			condlog(0, "cannot create " CALLOUT_DIR);
			FREE(buf);
			return -1;
		}
		condlog(4, "created " CALLOUT_DIR);
	}
	FREE(buf);

	/*
	 * compute the optimal ramdisk size
	 */
	vector_foreach_slot (conf->binvec, bin,i) {
		if ((fd = open(bin, O_RDONLY)) < 0) {
			if (errno == ENOENT)
				condlog(3, "callout %s does not exist, ignoring", bin);
			else
				condlog(0, "cannot open %s : %s", bin,
					strerror(errno));
			continue;
		}
		if (fstat(fd, &statbuf) < 0) {
			condlog(0, "cannot stat %s", bin);
			return -1;
		}
		size += statbuf.st_size;
		close(fd);
	}
	condlog(3, "ramfs maxsize is %u", (unsigned int) size);

	/*
	 * mount the ramfs
	 */
	if (safe_sprintf(ramfs_args, "maxsize=%u", (unsigned int) size)) {
		condlog(0, "ramfs_args too small\n");
		return -1;
	}
	if (mount(NULL, CALLOUT_DIR, "ramfs", MS_SYNCHRONOUS, ramfs_args) < 0) {
		condlog(0, "cannot mount ramfs on " CALLOUT_DIR);
		return -1;
	}
	condlog(4, "mount ramfs on " CALLOUT_DIR);

	/*
	 * populate the ramfs with callout binaries
	 */
	vector_foreach_slot (conf->binvec, bin,i) {
		if (copytodir(bin, CALLOUT_DIR) < 0)
			continue;
		condlog(4, "cp %s in ramfs", bin);
	}
	free_strvec(conf->binvec);
	conf->binvec = NULL;

	if (unmount_extra_devs() < 0)
		return -1;

	/*
	 * bind the ramfs to :
	 * /sbin : default home of multipath ...
	 * /bin  : default home of scsi_id ...
	 * /tmp  : home of scsi_id temp files
	 */
	if (mount(CALLOUT_DIR, "/sbin", NULL, MS_BIND, NULL) < 0) {
		condlog(0, "cannot bind ramfs on /sbin");
		return -1;
	}
	condlog(4, "bind ramfs on /sbin");
	if (mount(CALLOUT_DIR, "/bin", NULL, MS_BIND, NULL) < 0) {
		condlog(0, "cannot bind ramfs on /bin");
		return -1;
	}
	condlog(4, "bind ramfs on /bin");
	if (mount(CALLOUT_DIR, "/tmp", NULL, MS_BIND, NULL) < 0) {
		condlog(0, "cannot bind ramfs on /tmp");
		return -1;
	}
	condlog(4, "bind ramfs on /tmp");
	return 0;
}
#endif

static void *
signal_set(int signo, void (*func) (int))
{
	int r;
	struct sigaction sig;
	struct sigaction osig;

	sig.sa_handler = func;
	sigemptyset(&sig.sa_mask);
	sig.sa_flags = 0;

	r = sigaction(signo, &sig, &osig);

	if (r < 0)
		return (SIG_ERR);
	else
		return (osig.sa_handler);
}

static void
sighup (int sig)
{
	condlog(2, "reconfigure (SIGHUP)");

	lock(gvecs->lock);
	if (!exitting)
		reconfigure(gvecs);
	unlock(gvecs->lock);

#ifdef _DEBUG_
	dbg_free_final(NULL);
#endif
}

static void
sigend (int sig)
{
	exit_daemon(0);
}

static void
sigusr1 (int sig)
{
	condlog(3, "SIGUSR1 received");
}

static void
signal_init(void)
{
	signal_set(SIGHUP, sighup);
	signal_set(SIGUSR1, sigusr1);
	signal_set(SIGINT, sigend);
	signal_set(SIGTERM, sigend);
	signal(SIGPIPE, SIG_IGN);
}

static void
setscheduler (void)
{
        int res;
	static struct sched_param sched_param = {
		.sched_priority = 99
	};

        res = sched_setscheduler (0, SCHED_RR, &sched_param);

        if (res == -1)
                condlog(LOG_WARNING, "Could not set SCHED_RR at priority 99");
	return;
}

static void
set_oom_adj (int val)
{
	FILE *fp;

	fp = fopen("/proc/self/oom_adj", "w");

	if (!fp)
		return;

	fprintf(fp, "%i", val);
	fclose(fp);
}
	
static void
setup_daemon(void)
{
	int in_fd, out_fd;

	in_fd = open("/dev/null", O_RDONLY);
	if (in_fd < 0){
		fprintf(stderr, "cannot open /dev/null for input : %s\n",
			strerror(errno));
		_exit(0);
	}
	out_fd = open("/dev/console", O_WRONLY);
	if (out_fd < 0){
		fprintf(stderr, "cannot open /dev/console for output : %s\n",
			strerror(errno));
		_exit(0);
	}

	close(STDIN_FILENO);
	if (dup(in_fd) < 0) {
		fprintf(stderr, "cannot dup /dev/null to stdin : %s\n",
			strerror(errno));
		_exit(0);
	}
	close(STDOUT_FILENO);
	if (dup(out_fd) < 0) {
		fprintf(stderr, "cannot dup /dev/null to stdout : %s\n",
			strerror(errno));
		_exit(0);
	}
	close(STDERR_FILENO);
	if (dup(out_fd) < 0) {
		fprintf(stderr, "cannot dup /dev/null to stderr : %s\n",
			strerror(errno));
		_exit(0);
	}
	close(in_fd);
	close(out_fd);
	if (chdir("/") < 0)
		fprintf(stderr, "can't chdir to root directory\n");
}

enum child_flag_bits {
	__DAEMONIZE = 0,
#ifdef CLONE_NEWNS
	__NEW_NS,
#endif
};

#define DAEMONIZE	(1 << __DAEMONIZE)
#ifdef CLONE_NEWNS
#define NEW_NS		(1 << __NEW_NS)
#endif

static int
child (void * param)
{
	sigset_t old;
	size_t stacksize = 64 * 1024;
	pthread_t check_thr, uevent_thr, uxlsnr_thr;
	pthread_attr_t attr;
	struct vectors * vecs;
	unsigned int daemon = (unsigned long)param & DAEMONIZE;
#ifdef CLONE_NEWNS
	unsigned int new_ns = (unsigned long)param & NEW_NS;
#endif
	struct multipath * mpp;
	int i, ret;

	if (daemon)
		setup_daemon();
	mlockall(MCL_CURRENT | MCL_FUTURE);

	if (logsink)
		log_thread_start();

	condlog(2, "--------start up--------");
	condlog(2, "read " DEFAULT_CONFIGFILE);

	if (load_config(DEFAULT_CONFIGFILE))
		goto err;

	setlogmask(LOG_UPTO(conf->verbosity + 3));

	/*
	 * fill the voids left in the config file
	 */
	if (!conf->checkint) {
		conf->checkint = DEFAULT_CHECKINT;
		conf->max_checkint = MAX_CHECKINT(conf->checkint);
	}

	if (conf->max_fds) {
		struct rlimit fd_limit;
		
		fd_limit.rlim_cur = conf->max_fds;
		fd_limit.rlim_max = conf->max_fds;
		if (setrlimit(RLIMIT_NOFILE, &fd_limit) < 0)
			condlog(0, "can't set open fds limit to %d : %s\n",
				conf->max_fds, strerror(errno));
	}

	if (pidfile_create(DEFAULT_PIDFILE, getpid())) {
		condlog(0, "can not create pid file");
		goto err;
	}
	block_signal(SIGHUP, &old);
	signal_init();
	setscheduler();
	set_oom_adj(-16);
	vecs = gvecs = init_vecs();

	if (!vecs) {
		condlog(0, "can not allocate path vector");
		goto err;
	}

	if (sysfs_get_mnt_path(sysfs_path, FILE_NAME_SIZE)) {
		condlog(0, "can not find sysfs mount point");
		goto err;
	}
#ifdef CLONE_NEWNS
	if (new_ns && prepare_namespace() < 0) {
		condlog(0, "cannot prepare namespace");
		goto err;
	}
#endif

	/*
	 * fetch and configure both paths and multipaths
	 */
	if (configure(vecs, 1)) {
		condlog(0, "failure during configuration");
		goto err;
	}

	pthread_sigmask(SIG_SETMASK, &old, NULL);
	/*
	 * start threads
	 */
	pthread_attr_init(&attr);
	if (stacksize < PTHREAD_STACK_MIN)
		stacksize = PTHREAD_STACK_MIN;
	pthread_attr_setstacksize(&attr, stacksize);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
	
	pthread_create(&check_thr, &attr, checkerloop, vecs);
	pthread_create(&uevent_thr, &attr, ueventloop, vecs);
	pthread_create(&uxlsnr_thr, &attr, uxlsnrloop, vecs);

	pthread_cond_wait(&exit_cond, &exit_mutex);

	/*
	 * exit path
	 */
	block_signal(SIGHUP, NULL);
	lock(vecs->lock);
	exitting = 1;
	if (conf->queue_without_daemon == QUE_NO_DAEMON_OFF ||
	    conf->queue_without_daemon == QUE_NO_DAEMON_UNDEF)
		vector_foreach_slot(vecs->mpvec, mpp, i)
			dm_queue_if_no_path(mpp->alias, 0);
	remove_maps(vecs, stop_waiter_thread);
	free_pathvec(vecs->pathvec, FREE_PATHS);
	vecs->pathvec=NULL;

	pthread_cancel(check_thr);
	pthread_cancel(uevent_thr);
	pthread_cancel(uxlsnr_thr);

	/* It's not worth dealing with the races to free these
	free_keys(keys);
	keys = NULL;
	free_handlers(handlers);
	handlers = NULL;
	free_polls();
	*/

	ret = 0;
	while (num_waiters > 0 && ret == 0)
		ret = pthread_cond_wait(&waiter_cond, vecs->lock);
	if (ret) {
		condlog(0, "error while waiting for event threads: %s",
			strerror(errno));
		goto err;
	}
	unlock(vecs->lock);
	/* freeing vecs isn't worth the races
	pthread_mutex_destroy(vecs->lock);
	FREE(vecs->lock);
	vecs->lock = NULL;
	FREE(vecs);
	vecs = NULL;
	*/
	condlog(2, "--------shut down-------");
	
	if (logsink)
		log_thread_stop();

	dm_lib_release();
	dm_lib_exit();

	/*
	 * Freeing config must be done after condlog() and dm_lib_exit(),
	 * because logging functions like dlog() and dm_write_log()
	 * reference the config.
	 */
	free_config(conf);
	conf = NULL;

#ifdef _DEBUG_
	dbg_free_final(NULL);
#endif

	exit(0);
err:
	if(logsink)
		log_thread_stop();
	exit(1);
}

static int
daemonize(void)
{
	int pid;

	if( (pid = fork()) < 0){
		fprintf(stderr, "Failed first fork : %s\n", strerror(errno));
		return -1;
	}
	else if (pid != 0)
		return pid;

	setsid();

	if ( (pid = fork()) < 0){
		fprintf(stderr, "Failed second fork : %s\n", strerror(errno));
		_exit(1);
	}
	if (pid != 0)
		_exit(0);
#ifdef CLONE_NEWNS
	if (unshare(CLONE_NEWNS) < 0) {
		fprintf(stderr, "Failed to unshare namespace : %s\n", strerror(errno));
		return child((void *)DAEMONIZE);
	}
	return child((void *)(DAEMONIZE | NEW_NS));
#else
	return child((void *)DAEMONIZE);
#endif
}

int
main (int argc, char *argv[])
{
	extern char *optarg;
	extern int optind;
	int arg;
	int err;
	
	logsink = 1;
	dm_init();

	if (getuid() != 0) {
		fprintf(stderr, "need to be root\n");
		exit(1);
	}

	/* make sure we don't lock any path */
	if (chdir("/") < 0)
		fprintf(stderr, "can't chdir to root directory\n");
	umask(umask(077) | 022);

	conf = alloc_config();

	if (!conf)
		exit(1);

	while ((arg = getopt(argc, argv, ":dv:k::B")) != EOF ) {
	switch(arg) {
		case 'd':
			logsink = 0;
			//debug=1; /* ### comment me out ### */
			break;
		case 'v':
			if (sizeof(optarg) > sizeof(char *) ||
			    !isdigit(optarg[0]))
				exit(1);

			conf->verbosity = atoi(optarg);
			break;
		case 'B':
			conf->bindings_read_only = 1;
			break;
		case 'k':
			err = uxclnt(optarg);
			exit(err);
		default:
			;
		}
	}

	if (!logsink)
		return child((void *)0);
	else{
		err = daemonize();
	}
	
	if (err < 0)
		return 1;
	/* parent dies */
	return 0;
}

